/*
    Copyright  1995-2001, The AROS Development Team. All rights reserved.
    $Id: formatdate.c,v 1.6 2012/05/27 17:45:16 jacadcaps Exp $

    Desc:
    Lang: English
*/

#include "locale_intern.h"
#include <exec/types.h>
#include <utility/hooks.h>
#include <utility/date.h>
#define __NOLIBBASE__
#include <proto/utility.h>
#undef __NOLIBBASE__
#include <clib/alib_protos.h>

#include <stdio.h>

#include "formatdate.h"

static VOID PrintDigits(struct LocaleBase *LocaleBase, UWORD number, char fill, UWORD len, struct Hook *hook, struct Locale *locale);
static VOID _WriteChar(char token, CONST struct Hook *hook, CONST struct Locale *locale);
static VOID _WriteString(struct LocaleBase *LocaleBase, CONST_STRPTR string, struct Hook *hook, struct Locale *locale, LONG cmod);

static VOID _FormatDate(struct Locale *locale,
                 CONST_STRPTR formatString,
                 struct DateStamp *date,
                 struct ClockData *cData,
                 struct Hook *hook,
                 struct LocaleBase *LocaleBase, LONG def_fillchar, LONG def_cmod);

const ULONG dayspermonth[13] = {0 /* not used */,0,31,59,90,120,151,181,212,243,273,304,334};

/*****************************************************************************

    NAME */
#include <proto/locale.h>

	AROS_LH4(VOID, FormatDate,

/*  SYNOPSIS */
	AROS_LHA(struct Locale    *, locale, A0),
	AROS_LHA(STRPTR            , formatString, A1),
	AROS_LHA(struct DateStamp *, date, A2),
	AROS_LHA(struct Hook      *, hook, A3),

/*  LOCATION */
	struct LocaleBase *, LocaleBase, 10, Locale)

/* ---- locale.library/FormatDate
 *
 *   NAME
 *        FormatDate -- Create a date string using a formatting template.
 *
 *   FUNCTION
 *        Generate a date string based on a template. The bytes generated are sent
 *        to a user specified callback function.
 *
 *   INPUTS
 *
 *        locale        --  the locale to use when formatting the string
 *        formatString  --  the formatting template string; this is much like the
 *                          printf() formatting style, i.e. a % followed by a
 *                          formatting command. The following commands exist:
 *
 *
 *                    %a -- abbreviated weekday name
 *                    %A -- weekday name
 *                    %b -- abbreviated month name
 *                    %B -- month name
 *                    %c -- the same as "%a %b %d %H:%M:%S %Y"
 *                    %C -- the same as "%a %b %e %T %Y"
 *                    %d -- day number with leading zeros
 *                    %D -- the same as "%m/%d/%y"
 *                    %e -- day number with leading spaces
 *                    %F -- the same as "%Y-%m-%d"
 *                    %h -- abbreviated month name
 *                    %H -- hour using 24 hour style with leading zeros
 *                    %I -- hour using 12 hour style with leading zeros
 *                    %j -- julian date
 *                    %m -- month number with leading zeros
 *                    %M -- the number of minutes with leading zeros
 *                    %n -- linefeed
 *                    %p -- AM or PM string
 *                    %P -- am or pm string
 *                    %q -- hour using 24 hour style
 *                    %Q -- hour using 12 hour style
 *                    %r -- the same as "%I:%M:%S %p"
 *                    %R -- the same as "%H:%M"
 *                    %S -- the number of seconds with leading zeros
 *                    %t -- tab
 *                    %T -- the same as "%H:%M:%S"
 *                    %U -- the week number, taking Sunday as the first day
 *                          of the week
 *                    %w -- the weekday number
 *                    %W -- the week number, taking Monday as the first day
 *                          of the week
 *                    %x -- the same as "%m/%d/%y"
 *                    %X -- the same as "%H:%M:%S"
 *                    %y -- the year using two digits with leading zeros
 *                    %Y -- the year using four digits with leading zeros
 *              
 *        If the template parameter is NULL, a single null byte
 *        is sent to the callback function.
 *              
 *        date          --  the current date
 *        hook          --  callback function; this is called for every character
 *                          generated with the following arguments:
 *
 *                          * pointer to hook structure
 *                          * character
 *                          * pointer to locale
 *              
 *  SEE ALSO
 *        ParseDate(), <libraries/locale.h>
 * 
 * ----
 *             
 *  INTERNALS
 *           
 *  HISTORY            
 *  17.01.2000  bergers implemented U & W, j is still missing
 *  07.07.1999  SDuvan  implemented (U, W, j not done yet)
 */
{
    AROS_LIBFUNC_INIT

    struct ClockData cData;

    if (/* !locale || */ !hook)
	return;

	if (formatString)
	{
#warning Amiga2Date will fail around year 2114, because then the numer of seconds since 1978 dont fit in a 32 bit variable anymore!
   
		Amiga2Date(date->ds_Days * 86400 + date->ds_Minute * 60 + date->ds_Tick / 50, &cData);

		_FormatDate(locale, formatString, date, &cData, hook, LocaleBase, DEF_FILLCHAR, DEF_CASEMOD);
	}

	_WriteChar('\0', hook, locale);	/* Write null terminator */

	AROS_LIBFUNC_EXIT
} /* FormatDate */


#define FILLCHAR(y) (fillchar == FCHAR_UNSET ? (y) : fillchar)

static VOID _FormatDate(struct Locale *locale,
                 CONST_STRPTR formatString,
                 struct DateStamp *date,
                 struct ClockData *cData,
                 struct Hook *hook,
                 struct LocaleBase *LocaleBase, LONG def_fillchar, LONG def_cmod)
{
    ULONG week, days, tmp;
    UBYTE c;

    while ((c = *formatString++))
    {
    	if (c == '%')
		{
			LONG fillchar = def_fillchar, cmod = def_cmod;

	    c = *formatString++;

		switch (c)
		{
			case '_':
				fillchar = ' ';
				c = *formatString++;
				break;

			case '0':
				fillchar = '0';
				c = *formatString++;
				break;

			case '-':
				fillchar = FCHAR_NONE;
				c = *formatString++;
				break;

			case '^':
				cmod = CASEMOD_UPPER;
				c = *formatString++;
				break;
		}

		if (!c)
			break;

	    switch (c)
	    {
	    case 'a':
		_WriteString(LocaleBase, GetLocaleStr(locale, ABDAY_1 + cData->wday), hook, locale, cmod);
		break;
		
	    case 'A':
		_WriteString(LocaleBase, GetLocaleStr(locale, DAY_1 + cData->wday), hook, locale, cmod);
		break;
		
	    case 'b':
	    case 'h':
		_WriteString(LocaleBase, GetLocaleStr(locale, ABMON_1 + cData->month - 1), hook, locale, cmod);
		break;
		
	    case 'B':
		_WriteString(LocaleBase, GetLocaleStr(locale, MON_1 + cData->month - 1), hook, locale, cmod);
		break;
		
	    case 'c':
		_FormatDate(locale, "%a %b %d %H:%M:%S %Y", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'C':
		_FormatDate(locale, "%a %b %e %T %Y", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'd':
		PrintDigits(LocaleBase, cData->mday, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'x':
	    case 'D':
		_FormatDate(locale, "%m/%d/%y", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'e':
		PrintDigits(LocaleBase, cData->mday, FILLCHAR(' '), 2, hook, locale);
		break;

	    case 'F':
		_FormatDate(locale, "%Y-%m-%d", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'H':
		PrintDigits(LocaleBase, cData->hour, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'I':
		c = cData->hour;
		PrintDigits(LocaleBase, c > 12 ? c - 12 : c == 0 ? 12 : c, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'j':
#if 1
		/* 15-Aug-2003 bugfix: Now works like the original. - Piru
		 */
		{
			if (date)
			{
		    UWORD jul = date->ds_Days
		                - (cData->year * 365)
		                - (cData->year / 4)
		                + (cData->year / 100)
		                - (cData->year / 400) + 1;

		    PrintDigits(LocaleBase, jul,
		                FILLCHAR('0'),
		                3,
		                hook,
		                locale);
			}
			else
			{
				#warning I dont know how to fix this. - Ilkka
				PrintDigits(LocaleBase, cData->mday + dayspermonth[cData->month], FILLCHAR('0'), 3, hook, locale);
			}
		}
#else
	        /* TODO */
#warning Julian date not tested.
		/* Julian date is DDD (1 - 366)*/
		PrintDigits(LocaleBase,
		    cData->mday + dayspermonth[cData->month],
		    FILLCHAR('0'),
		    3,
		    hook,
		    locale
		);
#endif
		break;
		
	    case 'm':
		PrintDigits(LocaleBase, cData->month, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'M':
		PrintDigits(LocaleBase, cData->min, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'n':
		_WriteChar('\n', hook, locale);
		break;
		
	    case 'p':
		_WriteString(LocaleBase, GetLocaleStr(locale,
					  cData->hour < 12 ? AM_STR : PM_STR),
					  hook, locale, CASEMOD_NONE);
		break;

	    case 'P':
			_WriteString(LocaleBase, GetLocaleStr(locale, cData->hour < 12 ? AM_STR : PM_STR), hook, locale, CASEMOD_LOWER);
			break;

	    case 'q':
		PrintDigits(LocaleBase, cData->hour, FILLCHAR(-1), 2, hook, locale);
		break;
		
	    case 'Q':
		c = cData->hour;
		PrintDigits(LocaleBase, c > 12 ? c - 12 : c == 0 ? 12 : c, FILLCHAR(-1), 2, hook, locale);
		break;
		
	    case 'r':
#if 0
		_FormatDate(locale, "%I:%M:%S ", date, cData, hook, LocaleBase, fillchar, cmod);

		/* 15-Aug-2003 bugfix: Original is not using locale here, but
		 * built-in strings! However using locale makes much more
		 * sense here, so this "feature" is not reproduced here. - Piru
		 */
		_WriteString(cData->hour < 12 ? "AM" : "PM", hook, locale);
#else
		_FormatDate(locale, "%I:%M:%S %p", date, cData, hook, LocaleBase, fillchar, cmod);
#endif
		break;
		
	    case 'R':
		_FormatDate(locale, "%H:%M", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'S':
		PrintDigits(LocaleBase, cData->sec, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 't':
		_WriteChar('\t', hook, locale);
		break;
		
	    case 'X':
	    case 'T':
		_FormatDate(locale, "%H:%M:%S", date, cData, hook, LocaleBase, fillchar, cmod);
		break;
		
	    case 'W': /* week number, Monday first day of week */
	    case 'U': /* week number, Sunday first day of week */
	        days = cData->mday + dayspermonth[cData->month]; 
		
		/* leap year ? */
		if ((cData->year % 4) == 0 && cData->month > 2)
		{
		  /*
		  ** 1700, 1800, 1900, 2100, 2200 re not leap years.
		  ** 2000 is a leap year.
		  ** -> if a year is divisible by 100 but not by 400 then
		  ** it is not a leap year!
		  */
		  if ((cData->year % 100) == 0 && (cData->year % 400) != 0)
		    ;
		  else
		    days++;
		}
		
		/* 
		** If January 1st is a Monday then the first week
		** will start with a Sunday January 7th if Sunday is the first day of the week
		** but if Monday is the first day of the week then Jan 1st will also be the
		** first day of the first week.
		*/
		/* 
		** Go to Saturday = last day of week if Sunday is first day of week
                ** Go to Sunday   = last day of week if Monday is first day of week
		*/
		if (c == 'U')
		{
		  /* Sunday is first day of the week */
  		  tmp = days + (6 - cData->wday);
  		}
  		else
  		{
  		  /* Monday is first day of week */
  		  if (cData->wday != 0)
  		    tmp = days + (7 - cData->wday);
  		  else
  		    tmp = days;
		}

		if (tmp < 7)
		{
		  week = 0;
		}
		else
		{
		  /* cut off the few days that belong to week 0 */
		  tmp -= (tmp % 7);
		  /* Calculate the full amount of weeks */
		  week = tmp / 7;
		}
		
                PrintDigits(LocaleBase, week, FILLCHAR('0'), 2, hook, locale);
	    break;
		
	    case 'w':
		PrintDigits(LocaleBase, cData->wday, FILLCHAR(-1), 1, hook, locale);
		break;
		
	    case 'y':
		PrintDigits(LocaleBase, cData->year % 100, FILLCHAR('0'), 2, hook, locale);
		break;
		
	    case 'Y':
		PrintDigits(LocaleBase, cData->year, FILLCHAR('0'), 4, hook, locale);
		break;

            //case 'Z':
            //    /* current timezone Unimplemented in 3.1 */
            //    break;

	    default:
		_WriteChar(c, hook, locale);
		break;
	    }
	}
	else
	{
	    _WriteChar(c, hook, locale);
	}
    }
}


static VOID _WriteString(struct LocaleBase *LocaleBase, CONST_STRPTR string, struct Hook *hook, struct Locale *locale, LONG cmod)
{
    UBYTE c;

    while ((c = *string++))
    {
		if (cmod == CASEMOD_UPPER)
			c = ConvToUpper(locale, c);
		else if (cmod == CASEMOD_LOWER)
			c = ConvToLower(locale, c);

		_WriteChar(c, hook, locale);
    }
}


static VOID _WriteChar(char token, CONST struct Hook *hook, CONST struct Locale *locale)
{
     AROS_UFC3(VOID, hook->h_Entry,
       AROS_UFCA(CONST struct Hook *,   hook,   A0),
       AROS_UFCA(CONST struct Locale *, locale, A2),
       AROS_UFCA(char,            token,  A1)
     );
}


static VOID PrintDigits(struct LocaleBase *LocaleBase, UWORD number, char fill, UWORD len, struct Hook *hook,
		 struct Locale *locale)
{
    char  buf[7];
    char *ptr = &buf[6];
    int   i   = 0;
    
    buf[6] = 0;
    
    /*
     * 6-mar-2003: Fixed formatting of "0:10 am", it showed up as ":10 am".
     * - Piru
     *
     * while(number && i < len)
     */
    while ((number || !i) && i < len)
    {
	*--ptr = number % 10 + '0';
	number /= 10;
	i++;
    }

    while (len - i > 0  && (char)-1 != fill)
    {
      len--;
      _WriteChar(fill, hook, locale);
    }

    _WriteString(LocaleBase, (char *)ptr, hook, locale, CASEMOD_NONE);
}

VOID iFormatClockData(CONST struct Locale *locale, CONST_STRPTR formatString, CONST struct ClockData *cData, CONST struct Hook *hook, struct LocaleBase *LocaleBase)
{
	if (hook)
	{
		if (formatString)
		{
			_FormatDate((struct Locale *)locale, formatString, NULL, (struct ClockData *)cData, (struct Hook *)hook, LocaleBase, DEF_FILLCHAR, DEF_CASEMOD);
		}

		_WriteChar('\0', hook, locale);	/* Write null terminator */
	}
}
